import logging
from django.core.exceptions import ValidationError
from django.dispatch import receiver
from django.db.models import signals
from drfpasswordless.models import CallbackToken
from drfpasswordless.models import generate_numeric_token
from drfpasswordless.settings import api_settings
from drfpasswordless.services import TokenService
from drfpasswordless.utils import get_custom_user_model

logger = logging.getLogger(__name__)


@receiver(signals.post_save, sender=CallbackToken)
def invalidate_previous_tokens(sender, instance, created, **kwargs):
    """
    Invalidates all previously issued tokens of that type when a new one is created, used, or anything like that.
    """

    if instance.user.pk in api_settings.PASSWORDLESS_DEMO_USERS.keys():
        return

    if isinstance(instance, CallbackToken):
        CallbackToken.objects.active().filter(user=instance.user, type=instance.type).exclude(id=instance.id).update(is_active=False)


@receiver(signals.pre_save, sender=CallbackToken)
def check_unique_tokens(sender, instance, **kwargs):
    """
    Ensures that mobile and email tokens are unique or tries once more to generate.
    Note that here we've decided keys are unique even across auth and validation.
    We could consider relaxing this in the future as well by filtering on the instance.type.
    """
    if instance._state.adding:
        # save is called on a token to create it in the db
        # before creating check whether a token with the same key exists
        if isinstance(instance, CallbackToken):
            unique = False
            tries = 0

            if CallbackToken.objects.filter(key=instance.key, is_active=True).exists():
                # Try N(default=3) times before giving up.
                while tries < api_settings.PASSWORDLESS_TOKEN_GENERATION_ATTEMPTS:
                    tries = tries + 1
                    new_key = generate_numeric_token()
                    instance.key = new_key

                    if not CallbackToken.objects.filter(key=instance.key, is_active=True).exists():
                        # Leave the loop if we found a valid token that doesn't exist yet.
                        unique = True
                        break

                if not unique:
                    # A unique value wasn't found after three tries
                    raise ValidationError("Couldn't create a unique token even after retrying.")
            else:
                # A unique value was found immediately.
                pass

    else:
        # save is called on an already existing token to update it. Such as invalidating it.
        # in that case there is no need to check for the key. This way we both avoid an unneccessary db hit
        # and avoid to change key field of used tokens.
        pass


User = get_custom_user_model()


@receiver(signals.pre_save, sender=User)
def update_alias_verification(sender, instance, **kwargs):
    """
    Flags a user's email as unverified if they change it.
    Optionally sends a verification token to the new endpoint.
    """
    if isinstance(instance, User):

        if instance.id:

            if api_settings.PASSWORDLESS_USER_MARK_EMAIL_VERIFIED is True:
                """
                For marking email aliases as not verified when a user changes it.
                """
                email_field = api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME
                email_verified_field = api_settings.PASSWORDLESS_USER_EMAIL_VERIFIED_FIELD_NAME

                # Verify that this is an existing instance and not a new one.
                try:
                    user_old = User.objects.get(id=instance.id)  # Pre-save object
                    instance_email = getattr(instance, email_field)  # Incoming Email
                    old_email = getattr(user_old, email_field)  # Pre-save object email

                    if instance_email != old_email and instance_email != "" and instance_email is not None:
                        # Email changed, verification should be flagged
                        setattr(instance, email_verified_field, False)
                        if api_settings.PASSWORDLESS_AUTO_SEND_VERIFICATION_TOKEN is True:
                            email_subject = api_settings.PASSWORDLESS_EMAIL_VERIFICATION_SUBJECT
                            email_plaintext = api_settings.PASSWORDLESS_EMAIL_VERIFICATION_PLAINTEXT_MESSAGE
                            email_html = api_settings.PASSWORDLESS_EMAIL_VERIFICATION_TOKEN_HTML_TEMPLATE_NAME
                            message_payload = {'email_subject': email_subject,
                                               'email_plaintext': email_plaintext,
                                               'email_html': email_html}
                            success = TokenService.send_token(instance, 'email', CallbackToken.TOKEN_TYPE_VERIFY, **message_payload)

                            if success:
                                logger.info('drfpasswordless: Successfully sent email on updated address: %s'
                                            % instance_email)
                            else:
                                logger.info('drfpasswordless: Failed to send email to updated address: %s'
                                            % instance_email)

                except User.DoesNotExist:
                    # User probably is just initially being created
                    return

            if api_settings.PASSWORDLESS_USER_MARK_MOBILE_VERIFIED is True:
                """
                For marking mobile aliases as not verified when a user changes it.
                """
                mobile_field = api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME
                mobile_verified_field = api_settings.PASSWORDLESS_USER_MOBILE_VERIFIED_FIELD_NAME

                # Verify that this is an existing instance and not a new one.
                try:
                    user_old = User.objects.get(id=instance.id)  # Pre-save object
                    instance_mobile = getattr(instance, mobile_field)  # Incoming mobile
                    old_mobile = getattr(user_old, mobile_field)  # Pre-save object mobile

                    if instance_mobile != old_mobile and instance_mobile != "" and instance_mobile is not None:
                        # Mobile changed, verification should be flagged
                        setattr(instance, mobile_verified_field, False)
                        if api_settings.PASSWORDLESS_AUTO_SEND_VERIFICATION_TOKEN is True:
                            mobile_message = api_settings.PASSWORDLESS_MOBILE_MESSAGE
                            message_payload = {'mobile_message': mobile_message}
                            success = TokenService.send_token(instance, 'mobile', CallbackToken.TOKEN_TYPE_VERIFY, **message_payload)

                            if success:
                                logger.info('drfpasswordless: Successfully sent SMS on updated mobile: %s'
                                            % instance_mobile)
                            else:
                                logger.info('drfpasswordless: Failed to send SMS to updated mobile: %s'
                                            % instance_mobile)

                except User.DoesNotExist:
                    # User probably is just initially being created
                    pass



import logging
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import PermissionDenied
from django.core.validators import RegexValidator
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from drfpasswordless.models import CallbackToken
from drfpasswordless.settings import api_settings
from drfpasswordless.utils import (
    verify_user_alias,
    change_user_alias,
    validate_token_age,
    get_custom_user_model,
)

logger = logging.getLogger(__name__)
User = get_custom_user_model()
TOKEN_LENGTH = api_settings.PASSWORDLESS_CALLBACK_TOKEN_LENGTH


def clear_mobile_number(mobile):
    mobile = str(mobile)
    mobile = ''.join(mobile.split())
    mobile = ''.join(mobile.split('-'))
    mobile = ''.join(mobile.split('('))
    mobile = ''.join(mobile.split(')'))
    mobile = '+' + mobile if mobile[0] != '+' else mobile

    return mobile


class TokenField(serializers.CharField):
    default_error_messages = {
        'required': _('Token required'),
        'invalid': _('Invalid token'),
        'blank': _('Blank token'),
        'max_length': _('Tokens are {max_length} digits long.'),
        'min_length': _('Tokens are {min_length} digits long.')
    }


"""
Authentication
"""


class AbstractBaseAliasAuthenticationSerializer(serializers.Serializer):
    """
    Abstract class that returns a callback token based on the field given
    Returns a token if valid, None or a message if not.
    """

    @property
    def alias_type(self):
        # The alias type, either email or mobile
        raise NotImplementedError

    @property
    def alias_field_name(self):
        raise NotImplementedError

    def validate(self, attrs):
        alias = attrs.get(self.alias_type)

        if alias:
            # Create or authenticate a user
            # Return THem

            if api_settings.PASSWORDLESS_REGISTER_NEW_USERS is True:
                # If new aliases should register new users.
                try:
                    user = User.objects.get(
                        **{self.alias_field_name + '__iexact': alias}
                    )
                except User.DoesNotExist:
                    user = User.objects.create(**{self.alias_field_name: alias})
                    user.set_unusable_password()
                    user.save()
            else:
                # If new aliases should not register new users.
                try:
                    user = User.objects.get(
                        **{self.alias_field_name + '__iexact': alias}
                    )
                except User.DoesNotExist:
                    user = None

            if user:
                if not user.is_active:
                    # If valid, return attrs,
                    # so we can create a token in our logic controller.
                    msg = _('User account is disabled.')
                    raise serializers.ValidationError(msg)
            else:
                msg = _('No account is associated with this alias.')
                raise serializers.ValidationError(msg)
        else:
            msg = _('Missing %s.') % self.alias_type
            raise serializers.ValidationError(msg)

        attrs['user'] = user
        return attrs


class EmailAuthSerializer(AbstractBaseAliasAuthenticationSerializer):
    email = serializers.EmailField()

    @property
    def alias_type(self):
        return 'email'

    @property
    def alias_field_name(self):
        return api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME


class MobileAuthSerializer(AbstractBaseAliasAuthenticationSerializer):
    mobile = serializers.CharField(max_length=17)

    @property
    def alias_type(self):
        return 'mobile'

    @property
    def alias_field_name(self):
        return api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME

    def validate(self, attrs):
        if api_settings.PASSWORDLESS_MOBILE_NUMBER_STANDARDISE:
            attrs['mobile'] = clear_mobile_number(attrs['mobile'])

        phone_regex = RegexValidator(
            regex=r'^\+[1-9]\d{1,14}$',
            message="Mobile number must be entered in the format:"
                    " '+999999999'. Up to 15 digits allowed."
        )
        phone_regex(attrs['mobile'])

        return super().validate(attrs)


"""
Verification
"""


class AbstractBaseAliasVerificationSerializer(serializers.Serializer):
    """
    Abstract class that returns a callback token based on the field given
    Returns a token if valid, None or a message if not.
    """

    @property
    def alias_type(self):
        # The alias type, either email or mobile
        raise NotImplementedError

    @property
    def alias_field_name(self):
        raise NotImplementedError

    def validate(self, attrs):

        msg = _('There was a problem with your request.')

        if self.alias_type:
            # Get request.user
            # Get their specified valid endpoint
            # Validate

            request = self.context["request"]
            if request and hasattr(request, "user"):
                user = request.user
                if user:
                    if not user.is_active:
                        # If valid, return attrs,
                        # so we can create a token in our logic controller
                        msg = _('User account is disabled.')

                    else:
                        if hasattr(user, self.alias_field_name):
                            # Has the appropriate alias type
                            attrs['user'] = user
                            return attrs
                        else:
                            msg = _(
                                'This user doesn\'t have an %s.' % self.alias_field_name
                            )
            raise serializers.ValidationError(msg)
        else:
            msg = _('Missing %s.') % self.alias_type
            raise serializers.ValidationError(msg)


class EmailVerificationSerializer(AbstractBaseAliasVerificationSerializer):
    @property
    def alias_type(self):
        return 'email'

    @property
    def alias_field_name(self):
        return api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME


class MobileVerificationSerializer(AbstractBaseAliasVerificationSerializer):
    @property
    def alias_type(self):
        return 'mobile'

    @property
    def alias_field_name(self):
        return api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME


"""
Change
"""


class AbstractBaseAliasChangeSerializer(serializers.Serializer):
    """
    Abstract class that returns a callback token based on the field given
    Returns a token if valid, None or a message if not.
    """

    @property
    def alias_type(self):
        # The alias type, either email or mobile
        raise NotImplementedError

    @property
    def alias_field_name(self):
        raise NotImplementedError

    def validate(self, attrs):
        msg = _('There was a problem with your request.')

        if self.alias_type:
            # Get request.user
            # Get their specified valid endpoint
            # Validate

            request = self.context["request"]
            if request and hasattr(request, "user"):
                user = request.user
                if user:
                    if not user.is_active:
                        # If valid, return attrs,
                        # so we can create a token in our logic controller
                        msg = _('User account is disabled.')

                    else:
                        # Has the appropriate alias type
                        if hasattr(user, self.alias_field_name):
                            # Check, isn't alias the same.
                            alias = request.get(self.alias_type)
                            if getattr(user, self.alias_field_name) == alias:
                                msg = _(
                                    f'This user already have same {self.alias_type}.'
                                )
                            else:
                                attrs['user'] = user
                                return attrs
                        else:
                            msg = _(
                                'This user doesn\'t have an %s.' % self.alias_field_name
                            )
            raise serializers.ValidationError(msg)
        else:
            msg = _('Missing %s.') % self.alias_type
            raise serializers.ValidationError(msg)


class EmailChangeSerializer(AbstractBaseAliasVerificationSerializer):
    email = serializers.EmailField()

    @property
    def alias_type(self):
        return 'email'

    @property
    def alias_field_name(self):
        return api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME


class MobileChangeSerializer(AbstractBaseAliasVerificationSerializer):
    mobile = serializers.CharField(max_length=17)

    @property
    def alias_type(self):
        return 'mobile'

    @property
    def alias_field_name(self):
        return api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME

    def validate(self, attrs):
        if api_settings.PASSWORDLESS_MOBILE_NUMBER_STANDARDISE:
            attrs['mobile'] = clear_mobile_number(attrs['mobile'])

        phone_regex = RegexValidator(
            regex=r'^\+[1-9]\d{1,14}$',
            message="Mobile number must be entered in the format:"
                    " '+999999999'. Up to 15 digits allowed."
        )
        phone_regex(attrs['mobile'])
        return super().validate(attrs)


"""
Callback Token
"""


def token_age_validator(value):
    """
    Check token age
    Makes sure a token is within the proper expiration datetime window.
    """
    if api_settings.PASSWORDLESS_TEST_MODE:
        if int(value) in api_settings.PASSWORDLESS_TEST_CODE_INCORRECT:
            return False
        return True
    valid_token = validate_token_age(value)
    if not valid_token:
        raise serializers.ValidationError("The token you entered isn't valid.")
    return value


class AbstractBaseCallbackTokenSerializer(serializers.Serializer):
    """
    Abstract class inspired by DRF's own token serializer.
    Returns a user if valid, None or a message if not.
    """
    # Needs to be required=false to require both.
    email = serializers.EmailField(required=False)
    mobile = serializers.CharField(
        required=False,
        max_length=17
    )
    token = TokenField(
        min_length=TOKEN_LENGTH,
        max_length=TOKEN_LENGTH,
        validators=[token_age_validator]
    )

    def validate_alias(self, attrs):
        email = attrs.get('email', None)
        mobile = attrs.get('mobile', None)

        if email and mobile:
            raise serializers.ValidationError()

        if not email and not mobile:
            raise serializers.ValidationError()

        if email:
            return api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME, email
        elif mobile:
            return api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME, mobile

        return None

    def validate(self, attrs):
        if not attrs.get('mobile', None):
            return attrs

        if api_settings.PASSWORDLESS_MOBILE_NUMBER_STANDARDISE:
            attrs['mobile'] = clear_mobile_number(attrs['mobile'])

        phone_regex = RegexValidator(
            regex=r'^\+[1-9]\d{1,14}$',
            message="Mobile number must be entered in the format:"
                    " '+999999999'. Up to 15 digits allowed."
        )
        phone_regex(attrs['mobile'])
        return attrs


class CallbackTokenAuthSerializer(AbstractBaseCallbackTokenSerializer):

    def validate(self, attrs):
        # Check Aliases
        try:
            super().validate(attrs)
            alias_type, alias = self.validate_alias(attrs)
            callback_token = attrs.get('token', None)
            user = User.objects.get(**{alias_type + '__iexact': alias})

            if api_settings.PASSWORDLESS_TEST_MODE:
                if int(callback_token) in api_settings.PASSWORDLESS_TEST_CODE_INCORRECT:
                    raise serializers.ValidationError("Incorrect token from settings")

                token = CallbackToken(**{
                    'user': user,
                    'key': callback_token,
                    'type': CallbackToken.TOKEN_TYPE_AUTH,
                    'is_active': True,
                    'to_alias': alias,
                })
                if api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME == alias_type:
                    token.to_alias_type = 'MOBILE'
                elif api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME == alias_type:
                    token.to_alias_type = 'EMAIL'
                else:
                    raise serializers.ValidationError()
            else:
                token = CallbackToken.objects.get(**{
                    'user': user,
                    'key': callback_token,
                    'type': CallbackToken.TOKEN_TYPE_AUTH,
                    'is_active': True,
                })

            if token.user == user:
                # Check the token type for our uni-auth method.
                # authenticates and checks the expiry of the callback token.
                if not user.is_active:
                    msg = _('User account is disabled.')
                    raise serializers.ValidationError(msg)

                if (
                    api_settings.PASSWORDLESS_USER_MARK_EMAIL_VERIFIED or
                    api_settings.PASSWORDLESS_USER_MARK_MOBILE_VERIFIED
                ):
                    # Mark this alias as verified
                    user = User.objects.get(pk=token.user.pk)
                    success = verify_user_alias(user, token)

                    if success is False:
                        msg = _('Error validating user alias.')
                        raise serializers.ValidationError(msg)

                attrs['user'] = user
                return attrs

            else:
                msg = _('Invalid Token')
                raise serializers.ValidationError(msg)
        except CallbackToken.DoesNotExist:
            msg = _('Invalid alias parameters provided.')
            raise serializers.ValidationError(msg)
        except User.DoesNotExist:
            msg = _('Invalid user alias parameters provided.')
            raise serializers.ValidationError(msg)
        except ValidationError:
            msg = _('Invalid alias parameters provided.')
            raise serializers.ValidationError(msg)


class CallbackTokenVerificationSerializer(AbstractBaseCallbackTokenSerializer):
    """
    Takes a user and a token, verifies the token belongs to the user and
    validates the alias that the token was sent from.
    """

    def validate(self, attrs):
        try:
            super().validate(attrs)
            alias_type, alias = self.validate_alias(attrs)
            user_id = self.context.get("user_id")
            user = User.objects.get(**{'id': user_id, alias_type + '__iexact': alias})
            callback_token = attrs.get('token', None)

            if api_settings.PASSWORDLESS_TEST_MODE:
                if int(callback_token) in api_settings.PASSWORDLESS_TEST_CODE_INCORRECT:
                    raise serializers.ValidationError("Incorrect token from settings")

                token = CallbackToken(**{
                    'user': user,
                    'key': callback_token,
                    'type': CallbackToken.TOKEN_TYPE_VERIFY,
                    'is_active': True,
                    'to_alias': alias,
                })
                if api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME == alias_type:
                    token.to_alias_type = 'MOBILE'
                elif api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME == alias_type:
                    token.to_alias_type = 'EMAIL'
                else:
                    raise serializers.ValidationError()
            else:
                token = CallbackToken.objects.get(**{
                    'user': user,
                    'key': callback_token,
                    'type': CallbackToken.TOKEN_TYPE_VERIFY,
                    'is_active': True,
                })

            if token.user == user:
                # Mark this alias as verified
                success = verify_user_alias(user, token)
                if success is False:
                    logger.debug("drfpasswordless: Error verifying alias.")

                attrs['user'] = user
                return attrs
            else:
                msg = _('This token is invalid. Try again later.')
                logger.debug(
                    "drfpasswordless: User token mismatch when verifying alias.")

        except CallbackToken.DoesNotExist:
            msg = _('We could not verify this alias.')
            logger.debug("drfpasswordless: Tried to validate alias with bad token.")
            pass
        except User.DoesNotExist:
            msg = _('We could not verify this alias.')
            logger.debug("drfpasswordless: Tried to validate alias with bad user.")
            pass
        except PermissionDenied:
            msg = _('Insufficient permissions.')
            logger.debug("drfpasswordless: Permission denied while validating alias.")
            pass

        raise serializers.ValidationError(msg)


class CallbackTokenChangeSerializer(AbstractBaseCallbackTokenSerializer):
    """
    Takes a user and a token, verifies the token belongs to the user and
    validates the alias that the token was sent from.
    """

    def validate(self, attrs):
        try:
            super().validate(attrs)
            alias_type, alias = self.validate_alias(attrs)
            user_id = self.context.get('user_id', None)
            user = User.objects.get(**{'id': user_id})
            callback_token = attrs.get('token', None)
            if alias == getattr(user, alias_type):
                msg = _(
                    f'This user already have same {alias_type}.'
                )
                raise serializers.ValidationError(msg)

            if api_settings.PASSWORDLESS_TEST_MODE:
                if int(callback_token) in api_settings.PASSWORDLESS_TEST_CODE_INCORRECT:
                    raise serializers.ValidationError('Incorrect token from settings')

                token = CallbackToken(**{
                    'user': user,
                    'key': callback_token,
                    'type': CallbackToken.TOKEN_TYPE_CHANGE,
                    'is_active': True,
                    'to_alias': alias,
                })
                if api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME == alias_type:
                    token.to_alias_type = 'MOBILE'
                elif api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME == alias_type:
                    token.to_alias_type = 'EMAIL'
                else:
                    raise serializers.ValidationError()
            else:
                token = CallbackToken.objects.get(**{
                    'user': user,
                    'key': callback_token,
                    'type': CallbackToken.TOKEN_TYPE_CHANGE,
                    'is_active': True,
                })

            if token.user == user and token.to_alias == alias:
                # Change users alias and set as verified
                old_users = User.objects.filter(**{alias_type + '__iexact': alias})
                success = (
                    change_user_alias(user, token, old_users) and
                    verify_user_alias(user, token)
                )
                if success is False:
                    logger.debug("drfpasswordless: Error verifying alias.")

                attrs['user'] = user
                return attrs
            else:
                msg = _('This token is invalid. Try again later.')
                logger.debug(
                    "drfpasswordless: User token mismatch when verifying alias.")

        except CallbackToken.DoesNotExist:
            msg = _('We could not verify this alias. No token.')
            logger.debug("drfpasswordless: Tried to validate alias with bad token.")
            pass
        except User.DoesNotExist:
            msg = _('We could not verify this alias. Bad user.')
            logger.debug("drfpasswordless: Tried to validate alias with bad user.")
            pass
        except PermissionDenied:
            msg = _('Insufficient permissions.')
            logger.debug("drfpasswordless: Permission denied while validating alias.")
            pass

        raise serializers.ValidationError(msg)


"""
Responses
"""


class TokenResponseSerializer(serializers.Serializer):
    """
    Our default response serializer.
    """
    token = serializers.CharField(source='key')
    key = serializers.CharField(write_only=True)



# Generated by Django 3.0.2 on 2020-01-25 08:53

from django.db import migrations, models
import drfpasswordless.models


class Migration(migrations.Migration):

    dependencies = [
        ('drfpasswordless', '0003_callbacktoken_type'),
    ]

    operations = [
        migrations.AlterField(
            model_name='callbacktoken',
            name='key',
            field=models.CharField(default=drfpasswordless.models.generate_numeric_token, max_length=6),
        ),
        migrations.AlterUniqueTogether(
            name='callbacktoken',
            unique_together={('is_active', 'key', 'type')},
        ),
    ]



# Generated by Django 3.2.16 on 2022-12-29 05:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('drfpasswordless', '0006_auto_20221228_1353'),
    ]

    operations = [
        migrations.AlterField(
            model_name='callbacktoken',
            name='type',
            field=models.CharField(choices=[('AUTH', 'Auth'), ('VERIFY', 'Verify'), ('CHANGE', 'Change')], max_length=20),
        ),
    ]



import logging
import os
from django.core.exceptions import PermissionDenied
from django.core.mail import send_mail
from django.template import loader
from django.utils import timezone
from rest_framework.authtoken.models import Token
from drfpasswordless.models import CallbackToken
from drfpasswordless.settings import api_settings


logger = logging.getLogger(__name__)

TOKEN_LENGTH = api_settings.PASSWORDLESS_CALLBACK_TOKEN_LENGTH


def get_custom_user_model():
    """
    Return custom user model from PASSWORDLESS_USER_MODEL in settings,
    else return default user model
    """
    if api_settings.PASSWORDLESS_USER_MODEL:
        return api_settings.PASSWORDLESS_USER_MODEL
    else:
        return get_user_model()

User = get_custom_user_model()

def authenticate_by_token(callback_token):
    try:
        token = CallbackToken.objects.get(
            key=callback_token,
            is_active=True,
            type=CallbackToken.TOKEN_TYPE_AUTH
        )

        # Returning a user designates a successful authentication.
        token.user = User.objects.get(pk=token.user.pk)
        token.is_active = False  # Mark this token as used.
        token.save()

        return token.user

    except CallbackToken.DoesNotExist:
        logger.debug(
            "drfpasswordless: Challenged with a callback token that doesn't exist."
        )
    except User.DoesNotExist:
        logger.debug(
            "drfpasswordless: Authenticated user somehow doesn't exist."
        )
    except PermissionDenied:
        logger.debug(
            "drfpasswordless: Permission denied while authenticating."
        )

    return None


def create_callback_token_for_user(user, alias_type, token_type, alias=None):
    token = None
    alias_type_u = alias_type.upper()
    to_alias_field = getattr(
        api_settings,
        f'PASSWORDLESS_USER_{alias_type_u}_FIELD_NAME',
    )
    if user.pk in api_settings.PASSWORDLESS_DEMO_USERS.keys():
        token = CallbackToken.objects.filter(user=user).first()
        if token:
            return token
        else:
            return CallbackToken.objects.create(
                user=user,
                key=api_settings.PASSWORDLESS_DEMO_USERS[user.pk],
                to_alias_type=alias_type_u,
                to_alias=getattr(user, to_alias_field),
                type=token_type
            )

    if not alias:
        alias = getattr(user, to_alias_field)
    token = CallbackToken.objects.create(
        user=user,
        to_alias_type=alias_type_u,
        to_alias=alias,
        type=token_type
    )

    if token is not None:
        return token

    return None


def validate_token_age(callback_token):
    """
    Returns True if a given token is within the age expiration limit.
    """

    try:
        token = CallbackToken.objects.get(key=callback_token, is_active=True)
        seconds = (timezone.now() - token.created_at).total_seconds()
        token_expiry_time = api_settings.PASSWORDLESS_TOKEN_EXPIRE_TIME
        if token.user.pk in api_settings.PASSWORDLESS_DEMO_USERS.keys():
            return True
        if seconds <= token_expiry_time:
            return True
        else:
            # Invalidate our token.
            token.is_active = False
            token.save()
            return False

    except CallbackToken.DoesNotExist:
        # No valid token.
        return False


def verify_user_alias(user, token):
    """
    Marks a user's contact point as verified depending on accepted token type.
    """
    if token.to_alias_type == 'EMAIL':
        if token.to_alias == getattr(
            user,
            api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME,
        ):
            setattr(
                user,
                api_settings.PASSWORDLESS_USER_EMAIL_VERIFIED_FIELD_NAME,
                True,
            )
    elif token.to_alias_type == 'MOBILE':
        if token.to_alias == getattr(
            user,
            api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME,
        ):
            setattr(
                user,
                api_settings.PASSWORDLESS_USER_MOBILE_VERIFIED_FIELD_NAME,
                True,
            )
    else:
        return False
    user.save()
    return True


def change_user_alias(user, token, old_users):
    """
    Change a user's contact point depending on accepted token type.
    """
    if token.to_alias_type == 'EMAIL':
        setattr(
            user,
            api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME,
            token.to_alias,
        )
        old_users.update(**{
            api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME: "",
            api_settings.PASSWORDLESS_USER_EMAIL_VERIFIED_FIELD_NAME: False,
        })
    elif token.to_alias_type == 'MOBILE':
        setattr(
            user,
            api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME,
            token.to_alias,
        )
        old_users.update(**{
            api_settings.PASSWORDLESS_USER_MOBILE_FIELD_NAME: "",
            api_settings.PASSWORDLESS_USER_MOBILE_VERIFIED_FIELD_NAME: False,
        })
    else:
        return False
    user.save()
    return True


def inject_template_context(context):
    """
    Injects additional context into email template.
    """
    for processor in api_settings.PASSWORDLESS_CONTEXT_PROCESSORS:
        context.update(processor())
    return context


def send_email_with_callback_token(user, to_email, email_token, **kwargs):
    """
    Sends an Email to user.email.

    Passes silently without sending in test environment
    """

    try:
        if api_settings.PASSWORDLESS_EMAIL_NOREPLY_ADDRESS:
            # Make sure we have a sending address before sending.

            # Get email subject and message
            email_subject = kwargs.get(
                'email_subject',
                api_settings.PASSWORDLESS_EMAIL_SUBJECT,
            )
            email_plaintext = kwargs.get(
                'email_plaintext',
                api_settings.PASSWORDLESS_EMAIL_PLAINTEXT_MESSAGE,
            )
            email_html = kwargs.get(
                'email_html',
                api_settings.PASSWORDLESS_EMAIL_TOKEN_HTML_TEMPLATE_NAME,
            )

            # Inject context if user specifies.
            context = inject_template_context({'callback_token': email_token.key, })
            html_message = loader.render_to_string(email_html, context,)
            send_mail(
                email_subject,
                email_plaintext % email_token.key,
                api_settings.PASSWORDLESS_EMAIL_NOREPLY_ADDRESS,
                [to_email],
                fail_silently=False,
                html_message=html_message,
            )

        else:
            logger.debug(
                "Failed to send token email. "
                "Missing PASSWORDLESS_EMAIL_NOREPLY_ADDRESS."
            )
            return False
        return True

    except Exception as e:
        logger.debug(
            "Failed to send token email to user: %d."
            "Possibly no email on user object. Email entered was %s" %
            (user.id, getattr(user, api_settings.PASSWORDLESS_USER_EMAIL_FIELD_NAME))
        )
        logger.debug(e)
        return False


def send_sms_with_callback_token(user, to_number, mobile_token, **kwargs):
    """
    Sends an SMS to user.mobile via Twilio.

    Passes silently without sending in test environment.
    """
    if api_settings.PASSWORDLESS_TEST_SUPPRESSION is True:
        # we assume success to prevent spamming SMS during testing.
        # even if you have suppression on – you must provide a number
        # if you have mobile selected.
        if api_settings.PASSWORDLESS_MOBILE_NOREPLY_NUMBER is None:
            return False

        return True

    base_string = kwargs.get('mobile_message', api_settings.PASSWORDLESS_MOBILE_MESSAGE)

    try:
        if api_settings.PASSWORDLESS_MOBILE_NOREPLY_NUMBER:
            # We need a sending number to send properly

            from twilio.rest import Client
            twilio_client = Client(
                os.environ['TWILIO_ACCOUNT_SID'],
                os.environ['TWILIO_AUTH_TOKEN']
            )

            if to_number.__class__.__name__ == 'PhoneNumber':
                to_number = to_number.__str__()

            twilio_client.messages.create(
                body=base_string % mobile_token.key,
                to=to_number,
                from_=api_settings.PASSWORDLESS_MOBILE_NOREPLY_NUMBER
            )
            return True
        else:
            logger.debug(
                "Failed to send token sms. Missing PASSWORDLESS_MOBILE_NOREPLY_NUMBER."
            )
            return False
    except ImportError:
        logger.debug(
            "Couldn't import Twilio client. Is twilio installed?"
        )
        return False
    except KeyError:
        logger.debug(
            "Couldn't send SMS. "
            "Did you set your Twilio account tokens "
            "and specify a PASSWORDLESS_MOBILE_NOREPLY_NUMBER?"
        )
    except Exception as e:
        logger.debug(
            f"Failed to send token SMS to user: {user.id}. "
            f"Possibly no mobile number on user object "
            f"or the twilio package isn't set up yet. "
            f"Number entered was: "
            f"{to_number}"
        )
        logger.debug(e)
        return False


def create_authentication_token(user):
    """ Default way to create an authentication token"""
    return Token.objects.get_or_create(user=user)



import uuid
import string
from django.db import models
from django.conf import settings
from django.utils.crypto import get_random_string

from drfpasswordless.settings import api_settings


def generate_hex_token():
    return uuid.uuid1().hex


def generate_numeric_token():
    """
    Generate a random n-digit string of numbers.
    We use this formatting to allow leading 0s.
    """
    return get_random_string(
        length=api_settings.PASSWORDLESS_CALLBACK_TOKEN_LENGTH,
        allowed_chars=string.digits
    )


class CallbackTokenManger(models.Manager):
    def active(self):
        return self.get_queryset().filter(is_active=True)

    def inactive(self):
        return self.get_queryset().filter(is_active=False)


class AbstractBaseCallbackToken(models.Model):
    """
    Callback Authentication Tokens
    These tokens present a client with their authorization token
    on successful exchange of a random token (email) or token (for mobile)

    When a new token is created, older ones of the same type are invalidated
    via the pre_save signal in signals.py.
    """

    id = models.UUIDField(
        primary_key=True,
        default=uuid.uuid4,
        editable=False,
        unique=True,
    )
    created_at = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name=None,
        on_delete=models.CASCADE
    )
    is_active = models.BooleanField(default=True)
    to_alias = models.CharField(blank=True, max_length=254)
    to_alias_type = models.CharField(blank=True, max_length=20)

    objects = CallbackTokenManger()

    class Meta:
        abstract = True
        get_latest_by = 'created_at'
        ordering = ['-id']

    def __str__(self):
        return f'Key: {self.key}'


class CallbackToken(AbstractBaseCallbackToken):
    """
    Generates a random n-digit number to be returned.
    """
    TOKEN_TYPE_AUTH = 'AUTH'
    TOKEN_TYPE_VERIFY = 'VERIFY'
    TOKEN_TYPE_CHANGE = 'CHANGE'
    TOKEN_TYPES = (
        (TOKEN_TYPE_AUTH, 'Auth'),
        (TOKEN_TYPE_VERIFY, 'Verify'),
        (TOKEN_TYPE_CHANGE, 'Change'),
    )

    # key = models.CharField(default=generate_numeric_token, max_length=6)
    key = models.TextField(default=generate_numeric_token)
    type = models.CharField(max_length=20, choices=TOKEN_TYPES)

    def save(self, *args, **kwargs):
        """
        Check length for key.
        """
        if len(self.key) != api_settings.PASSWORDLESS_CALLBACK_TOKEN_LENGTH:
            raise ValueError(
                f'Key length not {api_settings.PASSWORDLESS_CALLBACK_TOKEN_LENGTH}'
            )
        super().save(*args, **kwargs)

    class Meta(AbstractBaseCallbackToken.Meta):
        verbose_name = 'Callback Token'



from django.contrib import admin
from django.urls import reverse
from drfpasswordless.models import CallbackToken


class UserLinkMixin(object):
    """
    A mixin to add a linkable list_display user field.
    """
    LINK_TO_USER_FIELD = 'link_to_user'

    def link_to_user(self, obj):
        link = reverse('admin:users_user_change', args=[obj.user.id])
        return u'<a href={}>{}</a>'.format(link, obj.user.username)
    link_to_user.allow_tags = True
    link_to_user.short_description = 'User'


class AbstractCallbackTokenInline(admin.StackedInline):
    max_num = 0
    extra = 0
    readonly_fields = ('created_at', 'key', 'type', 'is_active')
    fields = ('created_at', 'user', 'key', 'type', 'is_active')


class CallbackInline(AbstractCallbackTokenInline):
    model = CallbackToken


class AbstractCallbackTokenAdmin(UserLinkMixin, admin.ModelAdmin):
    readonly_fields = ('created_at', 'user', 'key', 'type', 'to_alias_type')
    list_display = ('created_at', UserLinkMixin.LINK_TO_USER_FIELD, 'key', 'type', 'is_active', 'to_alias_type')
    fields = ('created_at', 'user', 'key', 'type', 'is_active', 'to_alias_type')
    extra = 0
